;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 05_compiler.lisp - Kompilátor Scheme do zásobníkového jazyka
;;;

;;
;; Kompilaci provádí funkce compile-scheme, která očekává výraz Scheme
;; a vrací kód zásobníkového jazyka, jehož vykonání výraz vyhodnotí.
;; Například:
;; (compile-scheme '(let ((x 1)) (+ x 1)))
;;
;; Funkce compiled-evaluate vykoná zkompilovaný kód zadaného výrazu.
;; Jedná se o obdobu funkce evaluate, která vyhodnocuje zadaný výraz.
;; Například:
;; (compiled-evaluate '(let ((x 1)) (+ x 1)))
;; 
;; REPL, který používá kompilátor, se spouští výrazem:
;; (compiled-scheme-repl)
;;
;; Pro jednoduchost kompilaci provádíme v Lispu.
;; Výhodou je, že speciální operátory a makra můžeme psát pohodlněji v Lispu.


;;
;; Úložiště pro speciální operátory a makra
;;

;; Slovník speciálních operátorů.
;; Obsahuje páry (name . function), 
;; kde name je symbol určující jméno speciálního operátoru,
;; a function funkce zodpovědná za kompilaci speciálního operátoru.
;; Funkce je aplikována na seznam argumentů operátoru.
(defvar *specials* nil)

;; Vrátí funkci kompilující zadaný speciální operátor,
;; nebo nil v případě, že jméno není speciální operátor.
(defun find-special (name)
  (cdr (find name *specials* :key #'car)))
  
;; Slovník maker.
;; Obsahuje páry (name . expansion-function), 
;; kde name je symbol určující jméno makra,
;; a expansion-function expanzní funkce makra.
;; Funkce je aplikována na seznam argumentů makra.

(defvar *macros* nil)

;; Vrátí expanzní funkci zadaného makra,
;; nebo nil v případě, že jméno není makro.
(defun find-macro (name)
  (cdr (find name *macros* :key #'car)))

;;
;; Kompilování výrazu
;;

;; Kompilace výrazu
(defun compile-scheme (expr)
  (cond
   ((symbolp expr)
    `(,expr :val)) ; kompilace proměnné
   ((atom expr)
    `(,expr)) ; kompilace atomů, které nejsou proměnné
   (t
    (compile-comp expr)))) ; kompilace složeného výrazu

#|
(compile-scheme 'a)
(compile-scheme 1)
|#

;; Kompilace složeného výrazu
(defun compile-comp (expr)
  (let* ((op (car expr))
         (special (find-special op)))
    (if special
        (apply special (cdr expr)) ; kompilace speciálního operátoru
      (let ((macro (find-macro op)))
        (if macro
            (compile-scheme (apply macro (cdr expr))) ; kompilace makra
          (compile-app expr)))))) ; kompilace aplikace

;; Kompilace aplikace
(defun compile-app (expr)
  (append
   (reduce
    #'append
    (mapcar #'compile-scheme (cdr expr))) ; kompilace argumentů
   (compile-scheme (car expr)) ; kompilace operátoru
   '(:clsub)))

#|
(compile-scheme '(+ 1 a))
(compile-scheme '(+ 1 (* 2 3)))
|#

;;
;; Vstupní bod Scheme
;;
  
(defun compiled-evaluate (expr)
  (apply #'execute (compile-scheme expr)))

#|
(compiled-evaluate '(+ 1 (* 2 3)))
|#

;;
;; Speciální operátory
;;

;; Makro def-scheme-special pro definici speciálních operátorů v Scheme.
;; Syntax: 
;; (def-scheme-special name params
;;    body1
;;    ...
;;    bodyn)
;;
;; - name je symbol
;; – params je lambda seznam
;; - bodyi jsou výrazy

 
(defmacro def-scheme-special (name params &body body)
  `(push (cons ',name (lambda ,params ,@body)) *specials*))    


(def-scheme-special set! (var val)
  `(,@(compile-scheme val) ,var :set!)) 

(def-scheme-special if (test then else)
  `(,@(compile-scheme test) 
    :if
    ,@(compile-scheme else)
    :else
    ,@(compile-scheme then)
    :then))

(def-scheme-special begin (&rest body)
  (if (null body)
      (compile-scheme 'nil))
  `(,@(reduce #'append
              (mapcar (lambda (expr)
                        `(,@(compile-scheme expr) :drop))
                      (butlast body)))
    ,@(compile-scheme (car (last body)))))

(def-scheme-special lambda (params &rest body)
  `((,@(reduce
        #'append
        (mapcar (lambda (param)
                  `(,param :bind))
                (reverse params)))
     ,@(compile-scheme `(begin ,@body))
     ,@(mapcar (lambda (param) 
                 (declare (ignore param))
                 :unbind) params)
     :ret)))

(def-scheme-special quote (arg)
  `(,arg))

(def-scheme-special define (var val)
  `(,@(compile-scheme val) ,var :bind ,var))

(def-scheme-special do-while (&rest body)
  `(:lbl ,@(compile-scheme `(begin ,@body)) :cjmp nil))
  
(def-scheme-special atomic (&rest body)
  `(:aon
    ,@(compile-scheme `(begin ,@body))
    :aoff))        

(def-scheme-special assert (condition)
  `(,@(compile-scheme `(lambda () ,condition))
    :a? :if :aon :clsub :aoff :else :clsub :then 
    :if ,condition :assert-error :else :drop nil :then))

(def-scheme-special await (condition)
  `(,@(compile-scheme `(lambda () ,condition)) :await nil))

(def-scheme-special ts (var)
  `(:aon ,var :val t ,var :set! :drop :aoff))

;;
;; Makra
;;

;; Makro def-scheme-macro pro definici maker v Scheme.
;; Syntax: 
;; (def-scheme-macro name params
;;    body1
;;    ...
;;    bodyn)
;;
;; - name je symbol
;; – params je lambda seznam
;; - bodyi jsou výrazy

(defmacro def-scheme-macro (name args &body body)
  `(push (cons ',name (lambda ,args ,@body)) *macros*))


(def-scheme-macro let (bindings &rest body)
  `((lambda ,(mapcar #'car bindings)
      ,@body)
    ,@(mapcar #'cadr bindings)))

(def-scheme-macro list (&rest exprs)
  (if exprs
      `(cons ,(car exprs)
             (list ,@(cdr exprs)))
    nil))

(def-scheme-macro while (&rest args)
  `(if ,(car args)
       (do-while
        ,@(cdr args)
        ,(car args))
     nil))

(def-scheme-macro dotimes (arg &rest body)
  (let ((var (car arg))
        (count-form (cadr arg))
        (count (gensym "COUNT")))
    `(let ((,var 0)
           (,count ,count-form))
       (while (< ,var ,count)
              ,@body
              (set! ,var (+ ,var 1))))))


(def-scheme-macro and (&rest args)
  (if args
      (if (cdr args)
          `(if ,(car args)
               (and ,@(cdr args))
             nil)
        (car args))
    t))

(def-scheme-macro or (&rest args)
  (if args
      (if (cdr args)
          (let ((result (gensym "RESULT")))
            `(let ((,result ,(car args)))
               (if ,result ,result (or ,@(cdr args)))))
        (car args))
    nil))

(def-scheme-macro co (&rest exprs)
 (let ((start (gensym "START"))
       (end (gensym "END"))
       (arg (gensym "ARG"))
       (exprs-count (length exprs)))
   `(let ((,start (sem 0))
          (,end (sem 0)))
      ,@(mapcar (lambda (expr)
                  `(process-run-procedure (lambda (,arg)
                                            (p ,start)
                                            ,expr
                                            (v ,end))
                                          nil))
                exprs)
      (dotimes (i ,exprs-count)
        (v ,start))
      (dotimes (i ,exprs-count)
        (p ,end)))))

(def-scheme-macro with-lock (lock &rest body)
  (let ((lock-expr (car lock))
        (lock-sym (gensym "LOCK")))
    `(let ((,lock-sym ,lock-expr))
       (p ,lock-sym)
       (let ((result (begin ,@body)))
         (v ,lock-sym)
         result))))

(def-scheme-macro when (condition &rest body)
  `(if ,condition
       (begin ,@body)
     nil))

(def-scheme-macro unless (condition &rest body)
  `(if ,condition
       nil
       (begin ,@body)))

;;
;; REPL -- Scheme
;;

(defun compiled-scheme-repl ()
  (loop (format t "~%compiled scheme > ") 
        (handler-case (print (compiled-evaluate (read)))
          (error (err) (format t "~%Chyba: ~a" err)))))


;; Zátěžové testy 
;; Vyhodnocujte v (compiled-scheme-repl)
#|
;; Scheme

;; Test výkonu
(let ((n 0))
 (dotimes (i 10000)
   (set! n (+ n 1)))
 n)

;; Test souběžného vykonávání
(co
 (dotimes (i 1000)
   (print 1))
 (dotimes (i 1000)
   (print 2)))

;; Test souběhu
(let ((n 0))
  (co
   (dotimes (i 1000)
     (set! n (+ n 1)))
   (dotimes (i 1000)
     (set! n (+ n 1))))
  n)

;; Test složeného atomického příkazu
(let ((n 0))
  (co
   (dotimes (i 1000)
     [(set! n (+ n 1))])
   (dotimes (i 1000)
     [(set! n (+ n 1))]))
  n)

;; Test operace test-and-set
(let ((n 0) 
      (lock nil))
  (co
   (dotimes (i 1000)
     (while (ts lock)
            nil)
     (set! n (+ n 1))
     (set! lock nil))
   (dotimes (i 1000)
     (while (ts lock)
            nil)
     (set! n (+ n 1))
     (set! lock nil)))
  n)

;; Test Dekkerova algoritmu
(let ((in1 nil) 
      (in2 nil) 
      (turn 1) 
      (n 0))
  (co
   (dotimes (i 1000)
     (print (list 1 i))
     (set! in1 t)
     (while in2
       (when (= turn 2)
         (set! in1 nil)
         (await (= turn 1))
         (set! in1 t)))
     (set! n (+ n 1))
     (set! turn 2)
     (set! in1 nil))
   (dotimes (i 1000)
     (print (list 2 i))
     (set! in2 t)
     (while in1
       (when (= turn 1)
         (set! in2 nil)
         (await (= turn 2))
         (set! in2 t)))
     (set! n (+ n 1))
     (set! turn 1)
     (set! in2 nil)))
  n)

;; Test atomického vyhodnocování podmínky tvrzení
(let ((a 0))
  (co
   (dotimes (i 1000)
     {(= a a)})
   (dotimes (i 1000)
     (set! a 1)
     (set! a 0))))

;; Porovnej předchozí s:
(let ((a 0))
  (co
   (dotimes (i 1000)
     (print (= a a)))
   (dotimes (i 1000)
     (set! a 1)
     (set! a 0))))

;; Test zámku
(let ((n 0)
      (l (lock)))
  (co
   (dotimes (i 1000)
     (with-lock (l)
       (set! n (+ n 1))))
   (dotimes (i 1000)
     (with-lock (l)
       (set! n (+ n 1)))))
  n)
|#